//
// XMM Science Analysis System
//
// Giuseppe Vacanti (cosine) 
// June 2000
//
#include "TimeOrderedTable.h"
#include <errstr.h>
#include <Tracer.h>
#include <SasDalUtils.h>

#include <Dal.h>

#include <utility>
#include <algorithm>
#include <functional>

bool operator<(const TimeOrderedTable::TimeRow & rhs,
	       const TimeOrderedTable::TimeRow & lhs) {
  PROC
  RETURN rhs.first < lhs.first;
}


template<class TimeSelector>
TimeOrderedTable::TimeOrderedTable(const Table *tab,
				   const TimeSelector & selector,
				   const std::string & timename)
{
    const ColumnData * col = tab->column(timename)->data();
    fqn_ = tab->column(timename)->qualifiedName();
    STime::errorHandling = STime::EXCEPTION;
    using namespace SasDalUtils;
    for(RowIdx i = 0; i < tab->rows(); ++i){
	std::string timestring(getStringColumnRow(col, i));
	if(selector(timestring)){
	    stimes_.push_back(std::make_pair(STime(timestring,i)));
	} else {
	catch(STime::InvalidTime e){
	    errstr << "Encouneterd a time string that could not be translated to a valid time, " 
		   << fqn_ << std::endl
		   << " Row = " << i << " contains: " << getStringColumnRow(col, i)
		   << std::endl;
	    if(tolerant)
		Sas::warning("InvalidTimeString");
	    else
		Sas::error("InvalidTimeString");
	}
    }
    
    if(stimes_.size() > 0){
	// The table must be sorted in time ascending order.
	TimeRow prev = stimes_[0];
	for(RowIdx i = 1; i < stimes_.size(); ++i){
	    if(prev.first > stimes_[i].first)
		errstr
	  << fqn_ << " is not sorted in time ascending order at row "
	  << i << " This row = " << stimes_[i].first << ", the previous row = "
	  << prev.first << Sas::error("TableNotSortedByTime");
      prev = stimes_[i];
    }
  } else
    Sas::error(fqn_+ " does not contain any valid time record.",
	       "TableNotSortedByTime");
  delete col;
  STime::errorHandling = STime::ERRORANDEXIT;  
  ENDPROC
}

TimeOrderedTable::Range TimeOrderedTable::row_range(const STime & start,
						    const STime & stop) const
{
  PROC
  if(stop < start)
    errstr << "Trying to access " << fqn_
	   << " with an invalid time range. Start = " << start
	   << " Stop = " << stop << Sas::error("InvalidTimeRange");

  typedef std::vector<TimeRow>::const_iterator ITR;
  ITR il = std::lower_bound(stimes_.begin(), stimes_.end(),
			    std::make_pair(start, 0));
  ITR iu = std::upper_bound(stimes_.begin(), stimes_.end(),
			    std::make_pair(stop, 0));

  RowIdx rlo, rhi;

  if((il == stimes_.begin() || il == stimes_.end()) ||
     (iu == stimes_.begin() || iu == stimes_.end())) {
    rlo = rhi = stimes_.size();
    errstr << fqn_ << " looking for rows with time in "
	   << "[" << start << ", " << stop << "], but the available range is "
	   << "[" << stimes_[0].first << ", "
	   << stimes_[stimes_.size() - 1].first << "]"
	   << Sas::warning("TimeOutOfRange");
  } else {
    rlo = (*il).second;
    rhi = (*iu).second;
  }
  
  RETURN std::make_pair(rlo, rhi);
}

TimeOrderedTable::Range TimeOrderedTable::rows_before(const STime & time) const
{
  PROC
  typedef std::vector<TimeRow>::const_iterator ITR;
  const std::pair<ITR, ITR> p = std::equal_range(stimes_.begin(),
						 stimes_.end(),
						 std::make_pair(time, 0));  
  ITR pf = p.first;
  ITR ps = p.second;
  RowIdx rlo, rhi;
  if(pf == stimes_.end() || pf == stimes_.begin())
    rlo = stimes_.size();
  else
    rlo = (*pf).second - 1;

  if(ps == stimes_.end() || ps == stimes_.begin())
    rhi = stimes_.size();
  else
    rhi = (*ps).second - 1;

  RETURN std::make_pair(rlo, rhi);
    
}

static bool MatchingRows(const TimeOrderedTable::TimeRow & rhs,
			 const TimeOrderedTable::TimeRow & lhs)
{
  PROC
  RETURN rhs.second == lhs.second;
}

bool TimeOrderedTable::row_has_time(const RowIdx & row) const
{
  PROC
  RETURN std::binary_search(stimes_.begin(), stimes_.end(),
			    std::make_pair(STime(), row),
			    MatchingRows);
}

const STime& TimeOrderedTable::time_of_row(const RowIdx & row) const
{
  PROC
  if(!row_has_time(row))
    errstr << fqn_ << " Row = " << row << " does not have a matching time." 
	   << Sas::error("MissingTime");
  RETURN stimes_[row].first;
}


std::pair<TimeOrderedTable::iterator, TimeOrderedTable::iterator> TimeOrderedTable::range(const STime & start, const STime & stop) {
  PROC
  Range rr = row_range(start, stop);
  RETURN std::make_pair(&stimes_[rr.first], &stimes_[rr.second]);
}
